<?php

namespace CleverReach\BusinessLogic\Sync;

use CleverReach\BusinessLogic\Entity\RecipientAttribute;
use CleverReach\BusinessLogic\Interfaces\Attributes;
use CleverReach\BusinessLogic\Utility\Helper;
use CleverReach\Infrastructure\ServiceRegister;

/**
 * Class AttributesSyncTask.
 *
 * @package CleverReach\BusinessLogic\Sync
 */
class AttributesSyncTask extends BaseSyncTask {
  const INITIAL_PROGRESS_PERCENT = 5;

  /**
   * Array of global attribute IDs on CleverReach.
   *
   * @var array
   */
  private $globalAttributesIdsFromCR;

  /**
   * Map of default global attributes.
   *
   * @var array[attribute_nameattribute_type]
   */
  private static $defaultAttributesMap = [
    'salutation' => 'text',
    'title' => 'text',
    'firstname' => 'text',
    'lastname' => 'text',
    'street' => 'text',
    'zip' => 'text',
    'city' => 'text',
    'company' => 'text',
    'state' => 'text',
    'country' => 'text',
    'birthday' => 'date',
    'phone' => 'text',
    'shop' => 'text',
    'customernumber' => 'text',
    'language' => 'text',
    'newsletter' => 'text',
  ];

  /**
   * String representation of object.
   *
   * @inheritdoc
   */
  public function serialize() {
    return serialize($this->globalAttributesIdsFromCR);
  }

  /**
   * Constructs the object.
   *
   * @inheritdoc
   */
  public function unserialize($serialized) {
    $this->globalAttributesIdsFromCR = unserialize($serialized);
  }

  /**
   * Runs task execution.
   *
   * @throws \CleverReach\Infrastructure\Exceptions\InvalidConfigurationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpAuthenticationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpCommunicationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpRequestException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\RefreshTokenExpiredException
   */
  public function execute() {
    $globalAttributes = $this->getGlobalAttributesIdsFromCleverReach();
    // After retrieving all global attributes set initial progress.
    $progressPercent = self::INITIAL_PROGRESS_PERCENT;
    $this->reportProgress($progressPercent);

    $attributesToSend = $this->getAllAttributes();

    // Calculate progress step after setting initially progress.
    $totalAttributes = count($attributesToSend);
    $progressStep = (100 - self::INITIAL_PROGRESS_PERCENT) / $totalAttributes;
    $i = 0;
    foreach ($attributesToSend as $attribute) {
      $i++;
      if (isset($globalAttributes[$attribute['name']])) {
        $attributeIdOnCR = $globalAttributes[$attribute['name']];
        $this->getProxy()->updateGlobalAttribute($attributeIdOnCR, $attribute);
      }
      else {
        $this->getProxy()->createGlobalAttribute($attribute);
      }

      $progressPercent += $progressStep;
      if ($i === $totalAttributes) {
        $this->reportProgress(100);
      }
      else {
        $this->reportProgress($progressPercent);
      }
    }
  }

  /**
   * Get global attributes IDs from CleverReach.
   *
   * @return array
   *   Array of global attribute IDs on CleverReach.
   *
   * @throws \CleverReach\Infrastructure\Exceptions\InvalidConfigurationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpAuthenticationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpCommunicationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpRequestException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\RefreshTokenExpiredException
   */
  private function getGlobalAttributesIdsFromCleverReach() {
    if (empty($this->globalAttributesIdsFromCR)) {
      $this->globalAttributesIdsFromCR = $this->getProxy()->getAllGlobalAttributes();
    }

    return $this->globalAttributesIdsFromCR;
  }

  /**
   * Return all global attributes formatted for sending to CleverReach.
   *
   * Example:
   * [
   *   [
   *     'name' => 'email',
   *     'type' => 'text',
   *   ],
   *   [
   *     'name' => 'birthday',
   *     'type' => 'date',
   *   ]
   * ]
   *
   * @return array
   *   Array of global attributes supported by integration.
   */
  private function getAllAttributes() {
    /** @var \CleverReach\BusinessLogic\Interfaces\Attributes $attributesService */
    $attributesService = ServiceRegister::getService(Attributes::CLASS_NAME);
    $integrationAttributes = $attributesService->getAttributes();
    $attributesToSend = [];

    $attributeNames = array_keys(self::$defaultAttributesMap);
    foreach ($attributeNames as $attributeName) {
      if (($attribute = Helper::getAttributeByName($integrationAttributes, $attributeName)) !== NULL) {
        $attributesToSend[] = $this->createAttributeForSend($attribute);
      }
    }

    return $attributesToSend;
  }

  /**
   * Return formatted attribute.
   *
   * @param \CleverReach\BusinessLogic\Entity\RecipientAttribute $attribute
   *   Attribute entity returned from AttributeService.
   *
   * @return array
   *   Array representation of attribute.
   */
  private function createAttributeForSend(RecipientAttribute $attribute) {
    $attributeToSend['name'] = $attribute->getName();
    $attributeToSend['type'] = $attribute->getType() ?: self::$defaultAttributesMap[$attribute->getName()];
    $attributeToSend['description'] = $attribute->getDescription();
    $attributeToSend['preview_value'] = $attribute->getPreviewValue();
    $attributeToSend['default_value'] = $attribute->getDefaultValue();

    return $attributeToSend;
  }

}
