<?php

namespace CleverReach\BusinessLogic\Sync;

use CleverReach\BusinessLogic\Interfaces\OrderItems;
use CleverReach\Infrastructure\ServiceRegister;

/**
 * Class CampaignOrderSync.
 *
 * @package CleverReach\BusinessLogic\Sync
 */
class CampaignOrderSync extends BaseSyncTask {
  const INITIAL_PROGRESS_PERCENT = 10;

  /**
   * Associative array [item_id => mailing_id].
   *
   * @var array
   */
  private $orderItemsIdMailingIdMap;

  /**
   * CampaignOrderSync constructor.
   *
   * @param array $orderItemsIdMailingIdMap
   *   Associative array where Order item ID is key and mailing id is value.
   */
  public function __construct(array $orderItemsIdMailingIdMap) {
    $this->orderItemsIdMailingIdMap = $orderItemsIdMailingIdMap;
  }

  /**
   * String representation of object.
   *
   * @inheritdoc
   */
  public function serialize() {
    return serialize($this->orderItemsIdMailingIdMap);
  }

  /**
   * Constructs the object.
   *
   * @inheritdoc
   */
  public function unserialize($serialized) {
    $this->orderItemsIdMailingIdMap = unserialize($serialized);
  }

  /**
   * Runs task execution.
   *
   * @throws \CleverReach\Infrastructure\Exceptions\InvalidConfigurationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpAuthenticationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpCommunicationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpRequestException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\RefreshTokenExpiredException
   */
  public function execute() {
    $this->reportProgress(self::INITIAL_PROGRESS_PERCENT);
    /** @var \CleverReach\BusinessLogic\Entity\OrderItems $orderItemService */
    $orderItemService = ServiceRegister::getService(OrderItems::CLASS_NAME);
    $orderItems = $orderItemService->getOrderItems(array_keys($this->orderItemsIdMailingIdMap));

    if (!empty($orderItems)) {
      $this->reportAlive();
      $this->setMailingIds($orderItems);
      $this->reportProgress(50);
      $this->updateRecipientWithOrderItemsInformation($orderItems);
    }

    $this->reportProgress(100);
  }

  /**
   * Iterate through passed OrderItems and sets mailing ID.
   *
   * @param \CleverReach\BusinessLogic\Entity\OrderItem[]|null $orderItems
   *   array of Order item object.
   */
  private function setMailingIds($orderItems) {
    foreach ($orderItems as $orderItem) {
      $mailingId = $this->orderItemsIdMailingIdMap[$orderItem->getOrderItemId()];

      if ($mailingId !== NULL) {
        $orderItem->setMailingId($mailingId);
      }
    }
  }

  /**
   * Update recipient with with purchase information.
   *
   * @param \CleverReach\BusinessLogic\Entity\OrderItem[] $orderItems
   *   array of OrderItem objects fetched from OrderItemsService.
   *
   * @throws \CleverReach\Infrastructure\Exceptions\InvalidConfigurationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpAuthenticationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpCommunicationException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\HttpRequestException
   * @throws \CleverReach\Infrastructure\Utility\Exceptions\RefreshTokenExpiredException
   */
  private function updateRecipientWithOrderItemsInformation($orderItems) {
    $firstItem = $this->getFirstOrderItem($orderItems);
    $recipientEmail = $firstItem ? $firstItem->getRecipientEmail() : NULL;

    $this->getProxy()->uploadOrderItems($recipientEmail, $orderItems);
  }

  /**
   * Returns first item from array.
   *
   * @param \CleverReach\BusinessLogic\Entity\OrderItem[] $orderItems
   *   array of OrderItem objects fetched from OrderItemsService.
   *
   * @return \CleverReach\BusinessLogic\Entity\OrderItem|null
   */
  private function getFirstOrderItem($orderItems) {
    foreach ($orderItems as $orderItem) {
      return $orderItem;
    }

    return NULL;
  }

}
